@@ -547,6 +547,46 @@ extension DataReader where S == ErasedSequence<[UInt8]> {
 		self.init(data: Self.data(readingFrom: descriptor))
 	}
 
+	private static func read(size: Int, from io: DispatchIO) async throws -> [UInt8]? {
+		#if PROFILING
+			let id = OSSignpostID(log: readLog)
+			os_signpost(.begin, log: readLog, name: "Read", signpostID: id, "Starting read")
+		#endif
+		return try await withCheckedThrowingContinuation { continuation in
+			var chunk = DispatchData.empty
+			io.read(offset: 0, length: size, queue: .main) { done, data, error in
+				guard error == 0 else {
+					continuation.resume(throwing: NSError(domain: NSPOSIXErrorDomain, code: Int(error)))
+					return
+				}
+
+				chunk.append(data!)
+
+				#if PROFILING
+					os_signpost(.event, log: readLog, name: "Read", signpostID: id, "Read %td bytes", data!.count)
+				#endif
+
+				if done {
+					if chunk.isEmpty {
+						#if PROFILING
+							os_signpost(.end, log: readLog, name: "Read", signpostID: id, "Ended final read")
+						#endif
+						continuation.resume(returning: nil)
+					} else {
+						#if PROFILING
+							os_signpost(.end, log: readLog, name: "Read", signpostID: id, "Ended read")
+						#endif
+						continuation.resume(
+							returning: [UInt8](unsafeUninitializedCapacity: chunk.count) { buffer, count in
+								_ = chunk.copyBytes(to: buffer, from: nil)
+								count = chunk.count
+							})
+					}
+				}
+			}
+		}
+	}
+
 	public static func data(readingFrom descriptor: CInt) -> S {
 		let stream = BackpressureStream(backpressure: CountedBackpressure(max: 16), of: [UInt8].self)
 		let io = DispatchIO(type: .stream, fileDescriptor: descriptor, queue: .main) { _ in
@@ -560,50 +600,11 @@ extension DataReader where S == ErasedSequence<[UInt8]> {
 			let readSize = sysconf(CInt(_SC_PAGESIZE)) * 16
 		#endif
 
-		Task {
-			while await withCheckedContinuation({ continuation in
-				#if PROFILING
-					let id = OSSignpostID(log: readLog)
-					os_signpost(.begin, log: readLog, name: "Read", signpostID: id, "Starting read")
-				#endif
-				var chunk = DispatchData.empty
-				io.read(offset: 0, length: readSize, queue: .main) { done, data, error in
-					guard error == 0 else {
-						stream.finish(throwing: NSError(domain: NSPOSIXErrorDomain, code: Int(error)))
-						continuation.resume(returning: false)
-						return
-					}
-
-					chunk.append(data!)
-
-					#if PROFILING
-						os_signpost(.event, log: readLog, name: "Read", signpostID: id, "Read %td bytes", data!.count)
-					#endif
-
-					if done {
-						if chunk.isEmpty {
-							#if PROFILING
-								os_signpost(.end, log: readLog, name: "Read", signpostID: id, "Ended final read")
-							#endif
-							stream.finish()
-							continuation.resume(returning: false)
-						} else {
-							#if PROFILING
-								os_signpost(.end, log: readLog, name: "Read", signpostID: id, "Ended read")
-							#endif
-							let chunk = [UInt8](unsafeUninitializedCapacity: chunk.count) { buffer, count in
-								_ = chunk.copyBytes(to: buffer, from: nil)
-								count = chunk.count
-							}
-							Task {
-								await stream.yield(chunk)
-								continuation.resume(returning: true)
-							}
-						}
-					}
-				}
-			}) {
+		stream.task {
+			while let chunk = try await Self.read(size: readSize, from: io) {
+				await stream.yield(chunk)
 			}
+			stream.finish()
 		}
 
 		return .init(sequence: stream)
@@ -668,14 +669,7 @@ public struct File: Sendable {
 	}
 
 	#if canImport(Darwin)
-		static let blocksize = {
-			var buffer = stat()
-			// FIXME: This relies on a previous chdir to the output directory
-			stat(".", &buffer)
-			return buffer.st_blksize
-		}()
-
-		func compressedData() async -> [UInt8]? {
+		func compressedData(blocksizeHint: UInt32? = nil) async -> [UInt8]? {
 			guard !looksIncompressible else {
 				return nil
 			}
@@ -683,7 +677,7 @@ public struct File: Sendable {
 			// There is no benefit on APFS to using transparent compression if
 			// the data is less than one allocation block.
 			let totalSize = self.data.map(\.count).reduce(0, +)
-			guard totalSize > Self.blocksize else {
+			guard totalSize > blocksizeHint ?? 0 else {
 				return nil
 			}
 
@@ -773,7 +767,7 @@ public struct File: Sendable {
 			}
 		}
 
-		func write(compressedData data: [UInt8], toDescriptor descriptor: CInt) -> Bool {
+		func write(compressedData data: [UInt8], toDescriptor descriptor: CInt, outputDescriptor output: CInt) -> Bool {
 			let uncompressedSize = self.data.map(\.count).reduce(0, +)
 			let attribute =
 				"cmpf".utf8.reversed()  // magic
@@ -793,7 +787,7 @@ public struct File: Sendable {
 				return false
 			}
 
-			let resourceForkDescriptor = open(name + _PATH_RSRCFORKSPEC, O_WRONLY | O_CREAT, 0o666)
+			let resourceForkDescriptor = openat(output, name + _PATH_RSRCFORKSPEC, O_WRONLY | O_CREAT, 0o666)
 			guard resourceForkDescriptor >= 0 else {
 				return false
 			}
@@ -1120,10 +1114,12 @@ public enum Files: StreamAperture {
 	public struct Options: Sendable {
 		public let compress: Bool
 		public let dryRun: Bool
+		public let output: CInt
 
-		public init(compress: Bool, dryRun: Bool) {
+		public init(compress: Bool, dryRun: Bool, output: CInt? = nil) {
 			self.compress = compress
 			self.dryRun = dryRun
+			self.output = output ?? AT_FDCWD
 		}
 	}
 
@@ -1172,6 +1168,13 @@ public enum Files: StreamAperture {
 			var hardlinks = [File.Identifier: (String, Task<Void, Error>)]()
 			var directories = [Substring: Task<Void, Error>]()
 
+			#if canImport(Darwin)
+				var buffer = statfs()
+				let blocksize = fstatfs(options.output, &buffer) == 0 ? buffer.f_bsize : nil
+			#else
+				let blocksize = UInt32?.none
+			#endif
+
 			for try await file in files {
 				await completion.waitForCompletions()
 
@@ -1196,7 +1199,7 @@ public enum Files: StreamAperture {
 				func setStickyBit(on file: File) {
 					if file.sticky {
 						measureFilesystemOperation(on: file, named: "chmod") {
-							warn(chmod(file.name, mode_t(file.mode)), "Setting sticky bit on")
+							warn(fchmodat(options.output, file.name, mode_t(file.mode), 0), "Setting sticky bit on")
 						}
 					}
 				}
@@ -1223,7 +1226,7 @@ public enum Files: StreamAperture {
 						}
 
 						measureFilesystemOperation(on: file, named: "link") {
-							warn(link(original, file.name), "linking")
+							warn(linkat(options.output, original, options.output, file.name, 0), "linking")
 						}
 					}
 					continue
@@ -1240,7 +1243,7 @@ public enum Files: StreamAperture {
 							}
 
 							measureFilesystemOperation(on: file, named: "symlink") {
-								warn(symlink(String(data: Data(file.data.map(Array.init).reduce([], +)), encoding: .utf8)!, file.name), "symlinking")
+								warn(symlinkat(String(data: Data(file.data.map(Array.init).reduce([], +)), encoding: .utf8)!, options.output, file.name), "symlinking")
 							}
 							setStickyBit(on: file)
 						}
@@ -1254,7 +1257,7 @@ public enum Files: StreamAperture {
 							}
 
 							measureFilesystemOperation(on: file, named: "mkdir") {
-								warn(mkdir(file.name, mode_t(file.mode & 0o777)), "creating directory at")
+								warn(mkdirat(options.output, file.name, mode_t(file.mode & 0o777)), "creating directory at")
 							}
 							setStickyBit(on: file)
 						}
@@ -1270,7 +1273,7 @@ public enum Files: StreamAperture {
 									let compressedData =
 										options.compress
 										? try! await compressionStream.addTask {
-											await file.compressedData()
+											await file.compressedData(blocksizeHint: blocksize)
 										}.result.get() : nil
 								#endif
 
@@ -1279,7 +1282,7 @@ public enum Files: StreamAperture {
 								}
 
 								let fd = measureFilesystemOperation(on: file, named: "open") {
-									open(file.name, O_CREAT | O_WRONLY, mode_t(file.mode & 0o777))
+									openat(options.output, file.name, O_CREAT | O_WRONLY, mode_t(file.mode & 0o777))
 								}
 								if fd < 0 {
 									warn(fd, "creating file at")
@@ -1294,7 +1297,7 @@ public enum Files: StreamAperture {
 
 								#if canImport(Darwin)
 									if let compressedData = compressedData,
-										file.write(compressedData: compressedData, toDescriptor: fd)
+										file.write(compressedData: compressedData, toDescriptor: fd, outputDescriptor: options.output)
 									{
 										return
 									}
@@ -1413,7 +1416,7 @@ extension AsyncSequence where Element: Sendable, AsyncIterator: Sendable, Self:
 				("s", "statistics", "Print statistics on completion."),
 				("v", "verbose", "Print xip file contents."),
 			]
-			static let version = "3.2"
+			static let version = "3.3"
 
 			var input: String?
 			var output: String?
@@ -1429,7 +1432,7 @@ extension AsyncSequence where Element: Sendable, AsyncIterator: Sendable, Self:
 					} + [option(name: nil, has_arg: 0, flag: nil, val: 0)]
 				repeat {
 					let result = getopt_long(CommandLine.argc, CommandLine.unsafeArgv, Self.options.map(\.flag).reduce("", +), options, nil)
-					if result < 0 {
+					guard result >= 0 else {
 						break
 					}
 					switch UnicodeScalar(UInt32(result)) {
@@ -1591,18 +1594,31 @@ extension AsyncSequence where Element: Sendable, AsyncIterator: Sendable, Self:
 			let handle: FileHandle
 			if let input = options.input {
 				handle = try FileHandle(forReadingFrom: URL(fileURLWithPath: input))
-				try handle.seekToEnd()
-				await statistics.setTotal(Int(try handle.offset()))
+				if #available(macOS 10.15.4, iOS 13.4, watchOS 6.2, tvOS 13.4, *) {
+					try handle.seekToEnd()
+					await statistics.setTotal(Int(try handle.offset()))
+				} else {
+					handle.seekToEndOfFile()
+					await statistics.setTotal(Int(handle.offsetInFile))
+				}
 				try handle.seek(toOffset: 0)
 			} else {
 				handle = FileHandle.standardInput
 			}
 
-			if let output = options.output {
-				guard chdir(output) == 0 else {
-					fputs("Failed to access output directory at \(output): \(String(cString: strerror(errno)))", stderr)
+			let output = options.output.flatMap {
+				let output = open($0, O_RDONLY | O_DIRECTORY)
+				guard output >= 0 else {
+					fputs("Failed to access output directory at \($0): \(String(cString: strerror(errno)))\n", stderr)
 					exit(EXIT_FAILURE)
 				}
+				return output
+			}
+
+			defer {
+				if let output {
+					close(output)
+				}
 			}
 
 			let file = DataReader.data(readingFrom: handle.fileDescriptor)
@@ -1614,7 +1630,10 @@ extension AsyncSequence where Element: Sendable, AsyncIterator: Sendable, Self:
 				}
 			}
 
-			for try await file in Unxip.makeStream(from: .xip(), to: .disk(), input: DataReader(data: input), nil, nil, .init(compress: options.compress, dryRun: options.dryRun)) {
+			for try await file in Unxip.makeStream(
+				from: .xip(), to: .disk(), input: DataReader(data: input),
+				nil, nil, .init(compress: options.compress, dryRun: options.dryRun, output: output))
+			{
 				await statistics.note(file)
 				if options.verbose {
 					print(file.name)
